package org.drools.examples.threading;

import java.io.InputStreamReader;
import java.io.Reader;
import java.util.LinkedList;
import java.util.List;
import java.util.ListIterator;

import org.drools.RuleBase;
import org.drools.RuleBaseFactory;
import org.drools.StatefulSession;
import org.drools.compiler.PackageBuilder;
import org.drools.examples.threading.facts.Occurrence;
import org.drools.examples.threading.facts.OccurrenceSettings;
import org.drools.examples.threading.facts.Summary;
import org.drools.rule.Package;

/**
 * This is a sample class to fire rules launch a rule package from a rule source file.
 * @author Tate Moore
 */
public class ThreadingExample implements Runnable {

	private static final int DEFAULT_THREADS = 5; 

	// Assuming a RuleBase instance is thread safe as stated in documentation. 
	private static RuleBase ruleBase = null;
	private static Boolean runSynchronized = false;

    public static final void main(final String[] args) throws Exception {

    	// FROM 4.0 DOC: 
    	//       "A Rule Base instance is thread safe, in the sense that 
    	//       you can have the one instance shared across threads in 
    	//       your application, which may be a web application, for 
    	//       instance. The most common operation on a rulebase is 
    	//       to create a new rule session; either stateful or stateless."
    	ruleBase = ThreadingExample.createRuleBase();

    	// Isn't threaded yet but good practice.
    	synchronized (runSynchronized) {
        	runSynchronized = true;			
		}
    	List<Thread> threadList = new LinkedList<Thread>();
    	for (int i = 1; i <= DEFAULT_THREADS; i ++)
    	{
    		Thread t = new Thread(new ThreadingExample());
    		threadList.add(t);
    		t.start();
    		// If you don't let it init you will get the already known
    		// array out of bounds exception so let the engine init. 
    		Thread.sleep(100);
    	}

    	// Wait for all the threads to stop.
    	while (threadList.size() > 0) {
    		Thread.sleep(1000);
    		for (ListIterator<Thread> iterator = threadList.listIterator(); iterator.hasNext(); ) {
    			Thread t = iterator.next();
    			if (!t.isAlive()) {
    				t.join();
    				System.out.println("Thread[" + t.getName() + "] completed and was removed from list." );
    				iterator.remove();
    				t = null;
    			}
    		}
    	}
    	// Don't let 'err' stream in next tests beat 'out' to the console.
    	System.err.flush();
    	System.out.println("Thread[" + Thread.currentThread().getName() + "] synchronized version of test completed.  There should be no error output above!");
    	System.out.flush();

    	// Should only be the main thread now but good practice.
    	synchronized (runSynchronized) {
        	runSynchronized = false;			
		}
    	for (int i = 1; i <= DEFAULT_THREADS; i ++)
    	{
    		Thread t = new Thread(new ThreadingExample());
    		threadList.add(t);
    		t.start();
    		// If you don't let it init you will get the already known
    		// array out of bounds exception so let the engine init. 
    		Thread.sleep(100);
    	}

    	// Wait for all the threads to stop.
    	while (threadList.size() > 0) {
    		Thread.sleep(1000);
    		for (ListIterator<Thread> iterator = threadList.listIterator(); iterator.hasNext(); ) {
    			Thread t = iterator.next();
    			if (!t.isAlive()) {
    				t.join();
    				System.out.println("Thread[" + t.getName() + "] finished and remove from list." );
    				iterator.remove();
    				t = null;
    			}
    		}
    	}
    	System.err.flush();
    	System.out.println("Thread[" + Thread.currentThread().getName() + "] completed.  There should be no other output!");
    	System.out.flush();
    }

	public void run() {
		
		boolean runThisThreadSynchronized = false;
		synchronized (runSynchronized) {
	    	if (runSynchronized.booleanValue()) {
	    		runThisThreadSynchronized = true;
	    	}
    	}
		
		// Create/initialize independent instances of all objects for 
		// use within the session.
        Summary summary = new Summary();
        summary.setNewCount(0);
        summary.setCurrentCount(0);
        OccurrenceSettings occurrenceSettings = new OccurrenceSettings();
        occurrenceSettings.setMaxOccurrences(1000);
        occurrenceSettings.setMinOccurrences(10);

        // Used for checking the accuracy of the rule.
        long maxCount = 0; 
        if (occurrenceSettings.getMinOccurrences() != 0) { 
        	maxCount = occurrenceSettings.getMaxOccurrences() / occurrenceSettings.getMinOccurrences();
        }
        // Loop thru the rule 1011 times 
        // (1000 max + 10 per award + 1 for off by one) 
        // for general edge cases as well. 
        for(int i = 0; i <= 1011; i++) {
        	
        	if (runThisThreadSynchronized) {
        		synchronized (ThreadingExample.ruleBase) {
        			fileAllRules(summary, occurrenceSettings, maxCount, i);
        		}
        	} else {
        		fileAllRules(summary, occurrenceSettings, maxCount, i);
        	}
        		
        }
    	System.out.println("Thread[" + Thread.currentThread().getName() + "] run completed.");
    }

    public static RuleBase createRuleBase() throws Exception {
    	
    	// NOTE: This RuleBase construction code copied from HelloWorld.java example.
    	
        //read in the source
        final Reader source = new InputStreamReader( ThreadingExample.class.getResourceAsStream( "Threading.drl" ) );

        final PackageBuilder builder = new PackageBuilder();

        //this will parse and compile in one step
        builder.addPackageFromDrl( source );
        
        // Check the builder for errors
        if ( builder.hasErrors() ) {
            System.out.println( builder.getErrors().toString() );
            throw new RuntimeException( "Unable to compile \"Threading.drl\".");
        }

        //get the compiled package (which is serializable)
        final Package pkg = builder.getPackage();

        //add the package to a RuleBase (deploy the rule package).
        final RuleBase ruleBase = RuleBaseFactory.newRuleBase();
        ruleBase.addPackage( pkg );
        return ruleBase;
        // End copied construction code...
    	
    }
    
    public static void fileAllRules(Summary summary,
    							    OccurrenceSettings occurrenceSettings,
    							    long maxCount,
    							    long count) {
    	
    	StatefulSession session = ruleBase.newStatefulSession();
        
//      session.addEventListener( new DebugAgendaEventListener() );
//      session.addEventListener( new DebugWorkingMemoryEventListener() );
//      
//      final WorkingMemoryFileLogger logger = new WorkingMemoryFileLogger( session );
//      logger.setFileName( "log/threading" );        

    	session.insert( summary );

    	// Create a occurrence configuration object to vary the rule if you want. 
    	session.insert(occurrenceSettings);
      
    	Occurrence occurrence = new Occurrence();
    	occurrence.setCount(count);
    	session.insert(occurrence);

    	long correctAnswer = 0;
    	if (occurrenceSettings.getMinOccurrences() != 0) {
    		correctAnswer =  (occurrence.getCount() / occurrenceSettings.getMinOccurrences()) - summary.getCurrentCount();
    		if ((summary.getCurrentCount()+correctAnswer) > maxCount) {
    			correctAnswer = maxCount - summary.getCurrentCount();
    		}
    	}

     	session.fireAllRules();				

	  	// Show rules that don't get the correct answer.
	  	if (summary.getNewCount() != correctAnswer) {
	      	System.err.println(
	      			"ERROR Thread[" + Thread.currentThread().getName() + "]" +
	      			" occurrence[" + occurrence.getCount() + "]" +
	      			" currentCount[" + summary.getCurrentCount() + "]" +
	      			" newCount[" + summary.getNewCount() + "]" +
	      			" correctAnswer[" + correctAnswer + "]" 
	      	);
	      	summary.setCurrentCount(summary.getCurrentCount() + correctAnswer);
	  	}
	  	else {
	  		summary.setCurrentCount(summary.getCurrentCount() + summary.getNewCount());
	  	}
	  	summary.setNewCount(0);

//  	logger.writeToDisk();
	  	session.dispose();
    }
}
